----------------------------------------------------------------------------------
-- Company:        CERN, BE-CO 
-- Engineer:       Carlos Gil Soriano
-- 
-- Create Date:    23:57:51 06/10/2012 
-- Design Name:    m25p32 wishbone access
-- Module Name:    m25p32_core - Behavioral 
-- Project Name:   Level Conversion Circuits
-- Target Devices: Spartan 6
-- Tool versions: 
-- Description:    This is the logic that governs the access to the m25p32 Non
--                 volatile Flash memory
--
-- Dependencies: 
--
-- Revision: 
-- Revision 0.01 - File Created
-- Additional Comments: 
--
----------------------------------------------------------------------------------
library IEEE;
library work;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;
use work.m25p32_pkg.ALL;
use work.m25p32_core_pkg.ALL;
use work.spi_master_pkg.ALL;

entity m25p32_core is
   generic(g_INST_LENGTH : NATURAL := c_INST_LENGTH;
           g_ADDR_LENGTH : NATURAL := c_ADDR_LENGTH;
           g_DATA_LENGTH : NATURAL := c_DATA_LENGTH);
   port (
           wb_rst_i     : in  STD_LOGIC;
           wb_clk       : in  STD_LOGIC;

           mosi_o       : out STD_LOGIC; 
           miso_i       : in  STD_LOGIC;  
           sclk_o       : out STD_LOGIC;  
           ss_n_o       : out STD_LOGIC; 

           inst_db_i    : in  STD_LOGIC_VECTOR (8*g_INST_LENGTH - 1 downto 0);
           addr_db_i    : in  STD_LOGIC_VECTOR (8*g_ADDR_LENGTH - 1 downto 0);
           data_db_i    : in  STD_LOGIC_VECTOR (8*g_DATA_LENGTH - 1 downto 0);

           CTR0_o       : out STD_LOGIC_VECTOR (r_CTR0'a_length - 1 downto 0);
           FMOH_i       : in  STD_LOGIC_VECTOR (r_FMOH'a_length - 1 downto 0)
        );
end m25p32_core;

architecture Behavioral of m25p32_core is


signal MEM_fsm        : t_m25p32_fsm;
signal MEM_fsm_d0     : t_m25p32_fsm;

signal s_CTR0         : r_CTR0;
signal s_CTR1         : r_CTR1;
signal s_CTR1_slv     : STD_LOGIC_VECTOR (r_CTR1'a_length - 1 downto 0);
signal s_FMOH         : r_FMOH;

--! Registers from spi_master_pkg
signal s_SPI0         : r_SPI0;
signal s_SPI0_slv     : STD_LOGIC_VECTOR (31 downto 0);
signal s_SPI1         : r_SPI1;
signal s_SPI1_slv     : STD_LOGIC_VECTOR (31 downto 0);
signal s_SPI2         : r_SPI2;
signal s_SPI2_slv     : STD_LOGIC_VECTOR (15 downto 0);

signal s_wr_rqt       : STD_LOGIC;

signal s_inst_m25p32_regs  : STD_LOGIC_VECTOR (8*g_INST_LENGTH - 1 downto 0);
signal s_addr_m25p32_regs  : STD_LOGIC_VECTOR (8*g_ADDR_LENGTH - 1 downto 0);
signal s_data_m25p32_regs  : STD_LOGIC_VECTOR (8*g_DATA_LENGTH - 1 downto 0);

signal s_inst_m25p32_core  : STD_LOGIC_VECTOR (8*g_INST_LENGTH - 1 downto 0);
signal s_addr_m25p32_core  : STD_LOGIC_VECTOR (8*g_ADDR_LENGTH - 1 downto 0);
signal s_data_m25p32_core  : STD_LOGIC_VECTOR (8*g_DATA_LENGTH - 1 downto 0);

signal s_inst_spi          : STD_LOGIC_VECTOR (8*g_INST_LENGTH - 1 downto 0);
signal s_addr_spi          : STD_LOGIC_VECTOR (8*g_ADDR_LENGTH - 1 downto 0);
signal s_data_spi          : STD_LOGIC_VECTOR (8*g_DATA_LENGTH - 1 downto 0);


begin

   CTR0_o       <= f_STD_LOGIC_VECTOR(s_CTR0);
   s_CTR1       <= f_CTR1(s_CTR1_slv);

   s_FMOH       <= f_FMOH(FMOH_i);
   s_SPI0_slv   <= f_STD_LOGIC_VECTOR(s_SPI0);
   s_SPI1_slv   <= f_STD_LOGIC_VECTOR(s_SPI1);
   s_SPI2       <= f_SPI2(s_SPI2_slv);

   s_inst_m25p32_regs  <= inst_db_i;
   s_addr_m25p32_regs  <= addr_db_i;
   s_data_m25p32_regs  <= data_db_i;

   inst_spi_master_core: spi_master_core
   generic map(
       g_inst_length => c_INST_LENGTH,
       g_addr_length => c_ADDR_LENGTH,
       g_data_length => c_DATA_LENGTH
   ) port map(
       rst_i       => wb_rst_i, 
       clk_i       => wb_clk, 

       inst_i      => s_inst_spi, 
       addr_i      => s_addr_spi, 
       data_i      => s_data_spi, 

       SPI0_i      => s_SPI0_slv,
       SPI1_i      => s_SPI1_slv,
       SPI2_o      => s_SPI2_slv,

       spi_mosi_o  => mosi_o,
       spi_miso_i  => miso_i,
       spi_clk_o   => sclk_o,
       spi_cs_n_o  => ss_n_o 
   );


   --! if s_FMOH.BOV, Buffer OVerride flag is set to one, the SPI buffer
   --! is loaded from the m25p32 instead than from the double buffer.
   s_inst_spi <= s_inst_m25p32_regs when s_CTR1.BOV = '0'
                   else s_inst_m25p32_core;

   s_addr_spi <= s_addr_m25p32_regs when s_CTR1.BOV = '0'
                   else s_addr_m25p32_core;

   s_data_spi <= s_data_m25p32_regs when s_CTR1.BOV = '0'
                   else s_data_m25p32_core;


   --! @brief This is the combinational process that sets up the correct
   --!        instruction field from m25p32_core. 
   p_inst_m25p32_core: process(s_CTR1.BOV)
   begin
      case MEM_fsm is 
         when S1_WREN               =>
            s_inst_m25p32_core    <= c_WREN_inst;
         when S4_WRDI               =>
            s_inst_m25p32_core    <= c_WRDI_inst;
         when others                =>
            s_inst_m25p32_core    <= (others => '0');
      end case;
   end process;

   p_out: process (wb_clk)
   begin
      if rising_edge(wb_clk) then
         if wb_rst_i = '1' then
            s_CTR0         <= c_CTR0_default;
            s_SPI0         <= c_SPI0_default;
            s_SPI1         <= c_SPI1_default;
         else
            MEM_fsm_d0     <= MEM_fsm;
            if s_FMOH.OPP = '1' then
               case s_FMOH.OP is
                  when ERS    =>
                  when RDP    =>
                  when WRP    =>
                  when RDSR   =>
                  when WRSR   =>
                     proc_WRSR_fsm (s_CTR0, s_CTR1,
                                    s_SPI0, s_SPI1, s_SPI2,
                                    MEM_fsm, MEM_fsm_d0);
                  when others =>
               end case;
            else
               s_CTR0.OPA     <= '0';
               s_CTR0.OPF     <= '0';
            end if;
         end if;
      else
      end if;

--      --! Code added to remove the glitch at the end of WRDI instruction
--      if (    s_SPI2.SENT_OP'event
--          and s_SPI2.SENT_OP = '1'
--          and MEM_fsm = S4_WRDI   ) then
--             s_SPI1 <= c_SPI1_default;
--      end if;

   end process;

end Behavioral;
