----------------------------------------------------------------------------------
-- Company:        CERN, BE-CO-HT
-- Engineer:       Carlos Gil Soriano
-- 
-- Create Date:    17:50:12 02/07/2012 
-- Design Name:    control_top
-- Module Name:    control_top - Behavioral 
-- Project Name:   i2c to wishbone access controller
-- Target Devices: Spartan 6
-- Tool versions: 
-- Description:    This is the module that controls the access to the wishbone modules
--
-- Dependencies: 
--
-- Revision: 
-- Revision 0.01 - File Created
-- Additional Comments: 
--
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;

entity control_top is
  generic(
           g_WISHBONE_MODULES   : INTEGER := 6;
           g_WISHBONE_MEMWIDTH  : INTEGER := 32;
           g_WISHBONE_ADDRWIDTH : INTEGER := 16
         );
  port (
         -- This is a wishbone master
         rst                    : in   STD_LOGIC;
         wb_master_i2c_clk      : in   STD_LOGIC;
         wb_master_i2c_stb_o    : out  STD_LOGIC;
         wb_master_i2c_cyc_o    : out  STD_LOGIC;
         wb_master_i2c_sel_o    : out  STD_LOGIC;
         wb_master_i2c_we_o     : out  STD_LOGIC;
         wb_master_i2c_data_i   : in   STD_LOGIC_VECTOR (31 downto 0);      
         wb_master_i2c_data_o   : out  STD_LOGIC_VECTOR (31 downto 0);
         wb_master_i2c_addr_o   : out  STD_LOGIC_VECTOR (3  downto 0);
         wb_master_i2c_ack_i    : in   STD_LOGIC;
         wb_master_i2c_rty_i    : in   STD_LOGIC;
         wb_master_i2c_err_i    : in   STD_LOGIC;
         -- Interruptions
         ind_wb_addr_i     : in   STD_LOGIC;
         inst_rd_i         : in   STD_LOGIC;
         inst_wr           : in   STD_LOGIC;
         -- Here we add all the wishbone ports needed
         wb_stb_o          : out  STD_LOGIC_VECTOR (g_WISHBONE_MODULES - 1   downto 0);
         wb_cyc_o          : out  STD_LOGIC_VECTOR (g_WISHBONE_MODULES - 1   downto 0);
         wb_sel_o          : out  STD_LOGIC_VECTOR (g_WISHBONE_MODULES - 1   downto 0);
         wb_we_o           : out  STD_LOGIC_VECTOR (g_WISHBONE_MODULES - 1   downto 0);
         wb_data_i         : in   STD_LOGIC_VECTOR (g_WISHBONE_MEMWIDTH*g_WISHBONE_MODULES -1 downto 0);      
         wb_data_o         : out  STD_LOGIC_VECTOR (g_WISHBONE_MEMWIDTH*g_WISHBONE_MODULES -1 downto 0);
         wb_addr_o         : out  STD_LOGIC_VECTOR (g_WISHBONE_ADDRWIDTH*g_WISHBONE_MODULES - 1 downto 0);
         wb_ack_i          : in   STD_LOGIC_VECTOR (g_WISHBONE_MODULES - 1 downto 0);
         wb_rty_i          : in   STD_LOGIC_VECTOR (g_WISHBONE_MODULES - 1 downto 0);
         wb_err_i          : in   STD_LOGIC_VECTOR (g_WISHBONE_MODULES - 1 downto 0)
       );
end control_top;

architecture Behavioral of control_top is


  constant CTR0_addr             : STD_LOGIC_VECTOR(3 downto 0) := X"0";
  constant CTR1_addr             : STD_LOGIC_VECTOR(3 downto 0) := X"1";
  constant OP_addr               : STD_LOGIC_VECTOR(3 downto 0) := X"2";

  --
  -- module_number value
  --
  -- It is used later on on this core. It is as follows:
  -- 1     i2c_slave_addr     
  -- 2     trigger_addr       
  -- 3     multiboot_addr     
  -- 4     EEPROM_manager_addr
  -- 5     wr_core_addr       
  -- 6     EEPROM_memory      
  --

  constant control_addr          : STD_LOGIC_VECTOR (15 downto 0) := X"0100";
  constant i2c_slave_addr        : STD_LOGIC_VECTOR (15 downto 0) := X"0200";
  constant trigger_addr          : STD_LOGIC_VECTOR (15 downto 0) := X"0300";
  constant multiboot_addr        : STD_LOGIC_VECTOR (15 downto 0) := X"0400";
  constant EEPROM_manager_addr   : STD_LOGIC_VECTOR (15 downto 0) := X"0500";
  constant wr_core_addr          : STD_LOGIC_VECTOR (15 downto 0) := X"0600";
  constant EEPROM_memory         : STD_LOGIC_VECTOR (15 downto 0) := X"1000";

  constant i2c_DRXA              : STD_LOGIC_VECTOR(3 downto 0) := X"4"; 
  constant i2c_DRXB              : STD_LOGIC_VECTOR(3 downto 0) := X"5";
  constant i2c_DTX               : STD_LOGIC_VECTOR(3 downto 0) := X"6";

  type FETCHER_fsm is (S0_IDLE, S1_I2C_WADDR_INT, S2_I2C_WADDR_FETCHED, S3_WB_DATA_READ, S4_I2C_WRITE_DATA, S5_DONE, S6_WB_ERROR);
  type rights is array(0 to g_WISHBONE_MODULES + 1) of STD_LOGIC(1 downto 0);

  signal CTRL_FETCHER_fsm        : FETCHER_fsm;
  signal module_right            : rights;

  -------------------------------------
  -- STATUS register
  -------------------------------------
  -- 16 bits
  -- Wishbone access: Read-write
  ------------------------------------
  -- BIT       NAME           Description
  --  0         GEN            General ENable
  --  1         EN0            ENable access to device 0 in memory map 
  --  2         EN1            ENable access to device 1 in memory map
  --  3         EN2            ENable access to device 2 in memory map
  --  4         EN3            ENable access to device 3 in memory map
  --  5         EN4            ENable access to device 4 in memory map
  --  6         EN5            ENable access to device 5 in memory map
  --  7         EN6            ENable access to device 6 in memory map
  --  8         EN7            ENable access to device 7 in memory map
  --  31-9      x              Reserved
  -------------------------------------
  signal CTR0: STD_LOGIC_VECTOR (15 downto 0);

  -------------------------------------
  -- STATUS register
  -------------------------------------
  -- 16 bits
  -- Wishbone access: Read-write
  ------------------------------------
  -- BIT       NAME           Description
  --  0         GWR            General write permissions
  --  1         GRD            General read  permissions
  --  2         WR0            WRite permissions in memory map 0
  --  3         RD0            ReaD permissions in memory map 0
  --  4         WR1            WRite permissions in memory map 1
  --  5         RD1            ReaD permissions in memory map 1
  --  6         WR2            WRite permissions in memory map 2
  --  7         RD2            ReaD permissions in memory map 2
  --  8         WR3            WRite permissions in memory map 3
  --  9         RD3            ReaD permissions in memory map 3
  --  10        WR4            WRite permissions in memory map 4
  --  11        RD4            ReaD permissions in memory map 4
  --  12        WR5            WRite permissions in memory map 5
  --  13        RD5            ReaD permissions in memory map 5
  --  14        WR6            WRite permissions in memory map 6
  --  15        RD6            ReaD permissions in memory map 6
  --  16        WR7            WRite permissions in memory map 7
  --  17        RD7            ReaD permissions in memory map 7
  --  31-18     x              Reserved
  -------------------------------------
  signal CTR1: STD_LOGIC_VECTOR (15 downto 0);

  -------------------------------------
  -- OPERATION register
  -------------------------------------
  -- 32 bits
  -- Wishbone access: Read-write
  ------------------------------------
  -- BIT       NAME           Description
  --  0         RQT            operation ReQuesTed
  --  1         DONE           operation DONE
  --  2         ERROR          operation ERROR
  --  3         WR             WRite into WAO
  --  4         RD             ReaD from WAO
  --  8-5       MOD            wishbone MODule accessed
  --  15-9      x              Reserved
  --  31-16     WAO            Wishbone Address in Operation
  -------------------------------------

  signal OP: STD_LOGIC_VECTOR(31 downto 0);

  -------------------------------------
  -- Data ReaD from WishBone register
  -------------------------------------
  -- 32 bits
  -- Wishbone access: Read-only
  ------------------------------------
  -- BIT       NAME           Description
  --  31-0     DRD            Data ReaD from WishBone
  -------------------------------------
  signal DRD_WB: STD_LOGIC_VECTOR(31 downto 0);

  -------------------------------------
  -- Data WRite to WishBone
  -------------------------------------
  -- 32 bits
  -- Wishbone access: Read-write
  ------------------------------------
  -- BIT       NAME           Description
  --  31-0     DWR            Data WRite to WishBone
  -------------------------------------
  signal DWR_WB: STD_LOGIC_VECTOR(31 downto 0);

  signal ind_address : STD_LOGIC_VECTOR(15 downto 0);

begin

  gen_rights: for i in 0 to g_WISHBONE_MODULES generate
    module_right(i) <= CTRL(3+2*i) & CTRL(2+2*i);
  end generate gen_rights;


  reg_proc:  process(wb_clk)

    procedure reset_control is
    begin
      CTR0                  <= (others => '0'); 
      CTR1                  <= (others => '0');
      OP                    <= (others => '0');
      wb_master_i2c_stb_o   <= '0'; 
      wb_master_i2c_cyc_o   <= '0'; 
      wb_master_i2c_sel_o   <= '0';
      wb_master_i2c_we_o    <= '0'; 
      wb_master_i2c_data_o  <= (others => '0'); 
      wb_master_i2c_addr_o  <= (others => '0');
      -- Here we put to low the signals that interface with the rest of slaves 
      for i in 0 to (g_WISHBONE_MODULES - 1) loop
        wb_stb_o(i)           <= '0'; 
        wb_cyc_o(i)           <= '0'; 
        wb_sel_o(i)           <= '0'; 
        wb_we_o(i)            <= '0'; 
        wb_data_o(i)          <= (others => '0'); 
        wb_addr_o(i)          <= (others => '0'); 
      end loop;
    end procedure;

    procedure read_i2c_reg_rqt(reg_addr: IN STD_LOGIC_VECTOR(3 downto 0) )  is
    begin
      wb_master_stb_o         <= '1';
      wb_master_cyc_o         <= '1';
      wb_master_we_o          <= '0';
      wb_master_addr_o        <= reg_addr;
    end procedure;

    procedure read_back_i2c is
    begin
      if (wb_master_i2c_ack_i = '1') then
        OP(31 downto 16)        <= wb_master_data_i(15 downto 0);
        CTRL_FETCHER_fsm        <= S2_I2C_WADDR_FETCHED;
      else
        OP                      <= X"00000004";
        CTRL_FETCHER_fsm        <= S6_WB_ERROR;
      end if;
      wb_master_stb_o           <= '0';
      wb_master_cyc_o           <= '0';
    end procedure;


    procedure read_wb_reg_rqt(module_number: INTEGER) is
    begin
      CTRL_FETCHER_fsm          <= S3_WB_DATA_READ;
      OP(4 downto 0)            <= "10001";
      OP(8 downto 5)            <= std_logic_vector(to_unsigned(module_number-1,4));
      wb_stb_o(module_number)       <= '1'; 
      wb_cyc_o(module_number)       <= '1'; 
      wb_sel_o(module_number)       <= '0'; 
      wb_we_o(module_number)        <= '0'; 
      wb_addr_o(g_WISHBONE_ADDRWIDTH*module_number downto g_WISHBONE_ADDRWIDTH*(module_number-1))      <= OP(31 downto 16); 
    end procedure;

    procedure check_address_rights is
    begin
      -- Check the read access
      case OP(31 downto 24) is
        when X"01" =>
          if (module_right(0)(0) == '1')then
            read_wb_reg_rqt(1);
          else
          end if;
        when X"02" =>
          if (module_right(1)(0) == '1')then
            read_wb_reg_rqt(2);
          else
          end if;
        when X"03" =>
          if (module_right(2)(0) == '1')then
            read_wb_reg_rqt(3);
          else
          end if;
        when X"04" =>
          if (module_right(3)(0) == '1')then
            read_wb_reg_rqt(4);
          else
          end if;
        when X"05" =>
          if (module_right(4)(0) == '1')then
            read_wb_reg_rqt(5);
          else
          end if;
        when X"06" =>
          if (module_right(5)(0) == '1')then
            read_wb_reg_rqt(6);
          else
          end if;
        when X"10" =>
          if (module_right(6)(0) == '1')then
            read_wb_reg_rq(7);
          else
          end if;
        when others =>
      end case  
    end procedure;

    procedure write_i2c_data(data: IN STD_LOGIC_VECTOR(31 downto 0); addr: IN STD_LOGIC_VECTOR(3 downto 0)) is
    begin
      wb_master_i2c_stb_o   <= '1'; 
      wb_master_i2c_cyc_o   <= '1'; 
      wb_master_i2c_sel_o   <= '0'; 
      wb_master_i2c_we_o    <= '1'; 
      wb_master_i2c_data_o  <= data; 
      wb_master_i2c_addr_o  <= addr; 
    end procedure;

    procedure wb_to_i2c is
    begin
      case OP(8 downto 5) is
        when X"02"  =>
          if (wb_ack_i(0) = '1') then
            DRD_WB                   <= wb_data_i(g_WISHBONE_MEMWIDTH*1 - 1 downto 0);
            write_i2c_data(wb_data_i(g_WISHBONE_MEMWIDTH*1 -1 downto 0), X"6");
            CTRL_FETCHER_fsm         <= S4_I2C_WRITE_DATA;
          else
            DRD_WB            <= (others => '0');
            OP(4 downto 0)    <= "00100"; 
            CTRL_FETCHER_fsm  <= S6_WB_ERROR;
          end if;
        when X"03"  =>
          if (wb_ack_i(1) = '1') then
            DRD_WB                   <= wb_data_i(g_WISHBONE_MEMWIDTH*2 -1 downto g_WISHBONE_MEMWIDTH*1);
            write_i2c_data(wb_data_i(g_WISHBONE_MEMWIDTH*2 -1 downto  g_WISHBONE_MEMWIDTH*1), X"6");
            CTRL_FETCHER_fsm         <= S4_I2C_WRITE_DATA;
          else
            DRD_WB            <= (others => '0');
            OP(4 downto 0)    <= "00100"; 
            CTRL_FETCHER_fsm  <= S6_WB_ERROR;
          end if;
        when X"04"  =>
          if (wb_ack_i(2) = '1') then
            DRD_WB                   <= wb_data_i(g_WISHBONE_MEMWIDTH*3 -1 downto g_WISHBONE_MEMWIDTH*2);
            write_i2c_data(wb_data_i(g_WISHBONE_MEMWIDTH*3 -1 downto g_WISHBONE_MEMWIDTH*2), X"6");
            CTRL_FETCHER_fsm         <= S4_I2C_WRITE_DATA;
          else
            DRD_WB            <= (others => '0');
            OP(4 downto 0)    <= "00100"; 
            CTRL_FETCHER_fsm  <= S6_WB_ERROR;
          end if;
        when X"05"  =>
          if (wb_ack_i(3) = '1') then
            DRD_WB                   <= wb_data_i(g_WISHBONE_MEMWIDTH*4 -1 downto g_WISHBONE_MEMWIDTH*3);
            write_i2c_data(wb_data_i(g_WISHBONE_MEMWIDTH*4 -1 downto g_WISHBONE_MEMWIDTH*3), X"6");
            CTRL_FETCHER_fsm         <= S4_I2C_WRITE_DATA;
          else
            DRD_WB            <= (others => '0');
            OP(4 downto 0)    <= "00100"; 
            CTRL_FETCHER_fsm  <= S6_WB_ERROR;
          end if;
        when X"06"  =>
          if (wb_ack_i(4) = '1') then
            DRD_WB                   <= wb_data_i(g_WISHBONE_MEMWIDTH*5 -1 downto g_WISHBONE_MEMWIDTH*4);
            write_i2c_data(wb_data_i(g_WISHBONE_MEMWIDTH*5 -1 downto  g_WISHBONE_MEMWIDTH*4), X"6");
            CTRL_FETCHER_fsm         <= S4_I2C_WRITE_DATA;
          else
            DRD_WB            <= (others => '0');
            OP(4 downto 0)    <= "00100"; 
            CTRL_FETCHER_fsm  <= S6_WB_ERROR;
          end if;
        when X"10"  =>
          if (wb_ack_i(5) = '1') then
            DRD_WB                   <= wb_data_i(g_WISHBONE_MEMWIDTH*6 -1 downto g_WISHBONE_MEMWIDTH*5);
            write_i2c_data(wb_data_i(g_WISHBONE_MEMWIDTH*6 -1 downto  g_WISHBONE_MEMWIDTH*5), X"6");
            CTRL_FETCHER_fsm         <= S4_I2C_WRITE_DATA;
          else
            DRD_WB            <= (others => '0');
            OP(4 downto 0)    <= "00100"; 
            CTRL_FETCHER_fsm  <= S6_WB_ERROR;
          end if;
        when others =>
          DRD_WB            <= (others => '0');
          OP(4 downto 0)    <= "00100"; 
          CTRL_FETCHER_fsm  <= S6_WB_ERROR;
      end case;
    end procedure;

  begin

    if rising_edge(wb_clk)then
      if rst = '1' then
        reset_control;
      elsif ind_wb_addr_i = '1' then

        if CRTL_FETCHER_fsm = S0_IDLE then
          CTRL_FETCHER_fsm <= S1_I2C_WADDR_INTI;
          read_i2c_reg_rqt(i2c_DRXA);
        else
        --TODO: report this error
        end if;

      else
        case CTRL_FETCHER is
          when S0_IDLE               =>
          -- Nothing to do
          when S1_I2C_WADDR_INT      =>
            -- Take care of the interrupt and perform DRXA read
            read_back_i2c;
          when S2_I2C_WADDR_FETCHED  =>
            check_address_rights;
          when S3_WB_DATA_READ       =>
            read_back_wb;
          when S4_I2C_WRITE_DATA     =>
            OP(4 downto 0)       <= "00010";
            CTRL_FETCHER_fsm     <= S5_DONE;
          when S5_DONE               =>
            OP                   <= (others => '0');
            CTRL_FETCHER_fsm     <= S0_IDLE;
          when S6_WB_ERROR           => 
            OP                   <= (others => '0');
            CTRL_FETCHER_fsm     <= S0_IDLE;
          when others =>
        end case;
      end if;

    else
    end if;

  end process;

end Behavioral;
