--------------------------------------------------------------------------------
-- Company:        CERN, BE-CO-HT 
-- Engineer:       Carlos Gil Soriano
-- 
-- Create Date:    11:56:55 10/25/2011 
-- Design Name:    I2C slave register HDL
-- Module Name:    i2c_regs - Behavioral 
-- Project Name:   CTDAH
-- Target Devices: Spartan 6
-- Tool versions:  
-- Description:    This is the HDL that holds all the register of the I2C slave
--                 core
--
-- The registers should be programmed in the following order:
--   1.- PRE  Sets up the prescaler register
--   2.- CTR0 Sets up the control register: enable, start, i2c address
--   3.- CTR1 Sets up the threshold for detection signals in RX and TX buffers
--
-- Revision: 
-- Revision 0.01 - File Created
-- Additional Comments: 
--
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;

entity i2c_regs is
   port (
      -- These are the registers offers to others modules of the FPGA
      wb_rst_i  : in   STD_LOGIC;
      wb_clk    : in   STD_LOGIC;
      wb_we_i   : in   STD_LOGIC;
      wb_stb_i  : in   STD_LOGIC;
      wb_cyc_i  : in   STD_LOGIC;
      wb_sel_i  : in   STD_LOGIC;
      wb_data_i : in   STD_LOGIC_VECTOR (31 downto 0);
      wb_data_o : out  STD_LOGIC_VECTOR (31 downto 0);
      wb_addr_i : in   STD_LOGIC_VECTOR (3  downto 0);
      wb_ack_o  : out  STD_LOGIC;
      wb_rty_o  : out  STD_LOGIC;
      wb_err_o  : out  STD_LOGIC;

      i2c_addr  : in   STD_LOGIC_VECTOR (6  downto 0);
      -- These are the registers that are offered to the i2c slave core
      PRE_o     : out  STD_LOGIC_VECTOR (15 downto 0);
      CTR0_o    : out  STD_LOGIC_VECTOR (15 downto 0);
      CTR1_o    : out  STD_LOGIC_VECTOR (15 downto 0);
      STA_i     : in   STD_LOGIC_VECTOR (15 downto 0);
      DRX0_i    : in   STD_LOGIC_VECTOR (7  downto 0);
      DRX1_i    : in   STD_LOGIC_VECTOR (7  downto 0);
      DRX2_i    : in   STD_LOGIC_VECTOR (7  downto 0);
      DRX3_i    : in   STD_LOGIC_VECTOR (7  downto 0);
      DRX4_i    : in   STD_LOGIC_VECTOR (7  downto 0);
      DRX5_i    : in   STD_LOGIC_VECTOR (7  downto 0);
      DTX0_o    : out  STD_LOGIC_VECTOR (7  downto 0);
      DTX1_o    : out  STD_LOGIC_VECTOR (7  downto 0);
      DTX2_o    : out  STD_LOGIC_VECTOR (7  downto 0);
      DTX3_o    : out  STD_LOGIC_VECTOR (7  downto 0);
      -- Controls for FIFO_dispatcher
      load_TX     : out  STD_LOGIC;
      tx_fifo_oen : out  STD_LOGIC 
   );
end i2c_regs;

architecture Behavioral of i2c_regs is

   -------------------------------------
   -- STATUS register
   -------------------------------------
   -- 16 bits
   -- Wishbone access: Read-only
   ------------------------------------
   -- BIT       NAME           Description
   --  0         EN             General ENable
   --  1         RST            General ReSeT
   --  2         RD_WRN_INST    Reserved
   --  3         A_RX          
   --  4         A_TX          
   --  8-5       x              Reserved
   --  15-9      i2c_sla_fsm    i2c fsm
   -------------------------------------
  constant  STA_addr    : STD_LOGIC_VECTOR(3 downto 0) := X"0";

   -------------------------------------
   -- PRESCALER register
   -------------------------------------
   -- 16 bits
   -- Wishbone access: Write-read
   ------------------------------------
   -- BIT       NAME           Description
   --  15-0      PRE            Prescaler value
   -------------------------------------
   constant  PRE_addr    : STD_LOGIC_VECTOR(3 downto 0) := X"1";

   -------------------------------------
   -- CTR0 register
   -------------------------------------
   -- 16 bits
   -- Wishbone access: Write-read
   ------------------------------------
   -- BIT       NAME           Description
   --  0         EN             general ENable
   --  1         RST            ReSeT
   --  2         PEN            Prescaler ENable
   --  5-3       x              Reserved
   --  7-6       BIA            Bytes Indirect Addressing
   --  14-8      A[6:0]         I2c address
   --  15        x              Reserved
   -------------------------------------
   constant  CTR0_addr   : STD_LOGIC_VECTOR(3 downto 0) := X"2";

   -------------------------------------
   -- CTR1 register
   -------------------------------------
   -- 16 bits
   -- Wishbone access: Read-only
   ------------------------------------
   -- BIT       NAME           Description
   --  7-0       RDS            fsm status ReaD Status
   --  15-8      WRS            fsm status WRite Status
   -------------------------------------
   constant  CTR1_addr   : STD_LOGIC_VECTOR(3 downto 0) := X"3";

   -------------------------------------
   -- DRXA register
   -------------------------------------
   -- 16 bits
   -- Wishbone access: Read-only
   ------------------------------------
   -- BIT       NAME           Description
   --  7-0       DRX0           Data RX register 0      
   --  15-8      DRX1           Data RX register 1
   --  23-16     DRX2           Data RX register 2
   --  31-24     DRX3           Data RX register 3
   -------------------------------------
   constant  DRXA_addr   : STD_LOGIC_VECTOR(3 downto 0) := X"4";
   
   -------------------------------------
   -- DRXB register
   -------------------------------------
   -- 16 bits
   -- Wishbone access: Read-only
   ------------------------------------
   -- BIT       NAME           Description
   --  7-0       DRX5           Data RX register 5 
   --  15-8      DRX6           Data RX register 6 
   --  31-16     x              Reserved
   -------------------------------------
   constant  DRXB_addr   : STD_LOGIC_VECTOR(3 downto 0) := X"5";

   -------------------------------------
   -- DTX register
   -------------------------------------
   -- 16 bits
   -- Wishbone access: Read-only
   ------------------------------------
   -- BIT       NAME           Description
   --  7-0       RDS            fsm status ReaD Status
   --  15-8      WRS            fsm status WRite Status
   -------------------------------------
   constant  DTX_addr    : STD_LOGIC_VECTOR(3 downto 0) := X"6";

signal wb_ack_s : STD_LOGIC;
signal wb_rty_s : STD_LOGIC;
signal wb_err_s : STD_LOGIC;

signal STA_s : STD_LOGIC_VECTOR  (15 downto 0);
signal PRE_s  : STD_LOGIC_VECTOR (15 downto 0);
signal CTR0_s : STD_LOGIC_VECTOR (7  downto 0);
signal CTR1_s : STD_LOGIC_VECTOR (15 downto 0);

signal DTX0_s : STD_LOGIC_VECTOR(7 downto 0);
signal DTX1_s : STD_LOGIC_VECTOR(7 downto 0);
signal DTX2_s : STD_LOGIC_VECTOR(7 downto 0);
signal DTX3_s : STD_LOGIC_VECTOR(7 downto 0);

signal load_TX_s     : STD_LOGIC;
signal tx_fifo_oen_s : STD_LOGIC;

begin

PRE_o        <= PRE_s;
CTR0_o(15)          <= '0';
CTR0_o(14 downto 8) <= i2c_addr;
CTR0_o(7  downto 0) <= CTR0_s;
CTR1_o       <= CTR1_s;
DTX0_o       <= DTX0_s;
DTX1_o       <= DTX1_s;
DTX2_o       <= DTX2_s;
DTX3_o       <= DTX3_s;
wb_ack_o     <= wb_ack_s;
wb_rty_o     <= wb_rty_s;
wb_err_o     <= wb_err_s;
load_TX      <= load_TX_s;
tx_fifo_oen  <= tx_fifo_oen_s;

process (wb_clk)
   begin
      if rising_edge(wb_clk) then
         if wb_rst_i = '1' then
            wb_data_o    <= (others => '0');
            wb_ack_s     <= '0';
            wb_rty_s     <= '0';
            wb_err_s     <= '0';
            -- This is the default maybe it is good to place the
            -- autoconfiguration here.
            PRE_s   <= (others => '0');
            CTR0_s  <= (others => '0');
            CTR1_s  <= (others => '0');
            DTX0_s  <= (others => '0');
            DTX1_s  <= (others => '0');
            DTX2_s  <= (others => '0');
            DTX3_s  <= (others => '0');
            load_TX_s     <= '0';
            tx_fifo_oen_s <= '0';
         elsif (wb_stb_i = '1' and wb_cyc_i = '1') then
            if (wb_ack_s or wb_rty_s or wb_err_s) = '1' then
               wb_ack_s   <= '0';
               wb_rty_s   <= '0';
               wb_err_s   <= '0';
            else
            case wb_we_i is 
            when '1' =>
               case wb_addr_i is         
               when PRE_addr =>
                  PRE_s      <= wb_data_i(15 downto 0);
                  wb_ack_s   <= '1';
                  wb_rty_s   <= '0';
                  wb_err_s   <= '0';
              when CTR0_addr => 
                  CTR0_s     <= wb_data_i(7  downto 0); 
                  wb_ack_s   <= '1';
                  wb_rty_s   <= '0';
                  wb_err_s   <= '0';
               when CTR1_addr => 
                  CTR1_s     <= wb_data_i(15  downto 0); 
                  wb_ack_s   <= '1';
                  wb_rty_s   <= '0';
                  wb_err_s   <= '0';
               when DTX_addr =>
                  DTX0_s     <= wb_data_i(7  downto 0); 
                  DTX1_s     <= wb_data_i(15 downto 8); 
                  DTX2_s     <= wb_data_i(23 downto 16); 
                  DTX3_s     <= wb_data_i(31 downto 24); 
                  load_TX_s  <= '1';
                  wb_ack_s   <= '1';
                  wb_rty_s   <= '0';
                  wb_err_s   <= '0';
               when others =>
                  wb_ack_s   <= '0';
                  wb_rty_s   <= '0';
                  wb_err_s   <= '1';
               end case;
            when others =>
               case wb_addr_i is
               when STA_addr => 
                  wb_data_o(15 downto 0)  <= STA_i; 
                  wb_data_o(31 downto 16) <= (others => '0');
                  wb_ack_s                <= '1';
                  wb_rty_s                <= '0';
                  wb_err_s                <= '0';
               when PRE_addr =>
                  wb_data_o(15 downto 0)  <= PRE_s;
                  wb_data_o(31 downto 16) <= (others => '0');
                  wb_ack_s                <= '1';
                  wb_rty_s                <= '0';
                  wb_err_s                <= '0';
               when CTR0_addr => 
                  wb_data_o(7  downto 0)  <= CTR0_s;
                  wb_data_o(14 downto 8)  <= i2c_addr;
                  wb_data_o(31 downto 15) <= (others => '0');
                  wb_ack_s                <= '1';
                  wb_rty_s                <= '0';
                  wb_err_s                <= '0';
               when CTR1_addr => 
                  wb_data_o(15 downto 0)  <= CTR1_s; 
                  wb_data_o(31 downto 16) <= (others => '0');
                  wb_ack_s                <= '1';
                  wb_rty_s                <= '0';
                  wb_err_s                <= '0';
               when DRXA_addr => 
                  wb_data_o(7  downto 0)  <= DRX0_i; 
                  wb_data_o(15 downto 8)  <= DRX1_i; 
                  wb_data_o(23 downto 16) <= DRX2_i; 
                  wb_data_o(31 downto 24) <= DRX3_i; 
                  wb_ack_s                <= '1';
                  wb_rty_s                <= '0';
                  wb_err_s                <= '0';
               when DRXB_addr => 
                  wb_data_o(7  downto 0)  <= DRX4_i; 
                  wb_data_o(15 downto 8)  <= DRX5_i; 
                  wb_data_o(31 downto 16) <= (others => '0');
                  wb_ack_s                <= '1';
                  wb_rty_s                <= '0';
                  wb_err_s                <= '0';
               when DTX_addr => 
                  wb_data_o(7  downto 0)  <= DTX0_s; 
                  wb_data_o(15 downto 8)  <= DTX1_s; 
                  wb_data_o(23 downto 16) <= DTX2_s; 
                  wb_data_o(31 downto 24) <= DTX3_s;
                  wb_ack_s                <= '1';
                  wb_rty_s                <= '0';
                  wb_err_s                <= '0';
               when others  =>
                  wb_ack_s   <= '0';
                  wb_rty_s   <= '0';
                  wb_err_s   <= '1';
               end case;
            end case;
            end if;
         elsif load_TX_s = '1' then
            load_TX_s      <= '0';
            tx_fifo_oen_s  <= '1';
          elsif tx_fifo_oen_s = '1'then
            tx_fifo_oen_s  <= '0';
         end if;
      end if;
end process;

end Behavioral;

