----------------------------------------------------------------------------------
-- Company:        CERN, BE-CO-HT 
-- Engineer:       Carlos Gil Soriano
-- 
-- Create Date:    11:29:56 10/25/2011 
-- Design Name:    I2C Slave to Wishbone bridge 
-- Module Name:    i2c_slave_top - Behavioral 
-- Project Name:   CTDAH
-- Target Devices: Spartan 6
-- Tool versions: 
-- Description:    This is a bridge from I2C to Wishbone
--
-- Dependencies:   i2c_debouncer.vhd
--
-- Revision: 
-- Revision 0.01 - File Created
-- Additional Comments: 
--
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;

entity i2c_slave_core is
   port (
      -- Wishbone port
      wb_clk    : in  STD_LOGIC;
      wb_rst_i  : in  STD_LOGIC;
      -- I2C pins
      sda_oen    : out  STD_LOGIC;
      sda_i     : in   STD_LOGIC;
      sda_o     : out  STD_LOGIC;
      scl_oen    : out  STD_LOGIC;
      scl_i     : in   STD_LOGIC;
      scl_o     : out  STD_LOGIC;
      -- Registers directly accesible
      PRE_i     : in   STD_LOGIC_VECTOR (15 downto 0);
      CTR0_i    : in   STD_LOGIC_VECTOR (15 downto 0);
      CTR1_i    : in   STD_LOGIC_VECTOR (15 downto 0);
      STA_o     : out  STD_LOGIC_VECTOR (15 downto 0);
      DRX0_o    : out  STD_LOGIC_VECTOR (7  downto 0);
      DRX1_o    : out  STD_LOGIC_VECTOR (7  downto 0);
      DRX2_o    : out  STD_LOGIC_VECTOR (7  downto 0);
      DRX3_o    : out  STD_LOGIC_VECTOR (7  downto 0);
      DRX4_o    : out  STD_LOGIC_VECTOR (7  downto 0);
      DRX5_o    : out  STD_LOGIC_VECTOR (7  downto 0);
      DTX0_i    : in   STD_LOGIC_VECTOR (7  downto 0);
      DTX1_i    : in   STD_LOGIC_VECTOR (7  downto 0);
      DTX2_i    : in   STD_LOGIC_VECTOR (7  downto 0);
      DTX3_i    : in   STD_LOGIC_VECTOR (7  downto 0);
      load_TX     : in   STD_LOGIC;
      tx_fifo_oen : in   STD_LOGIC; 
      -- Alarms for controlling the i2c states
      ind_wb_addr : out  STD_LOGIC;
      inst_rd     : out  STD_LOGIC;
      inst_wr     : out  STD_LOGIC 
);
end i2c_slave_core;

architecture Behavioral of i2c_slave_core is


-------------------------------------------------------------------------------
-- This is the declaration area of the constants used in other components:
-- c_PRESCALER        default value of the prescaler register
-- c_COUNTER_WIDTH    bit length of the gc_counter
-- c_DATA_WIDTH       width of the data to be stored in FIFO_stack 
--                    and FIFO_dispatcher.
-- c_RX_STACK_DEPTH   depth of FIFO_stack
-- c_TX_STACK_DEPTH   depth of FIFO_dispatch 
-------------------------------------------------------------------------------


constant c_PRESCALER         : STD_LOGIC_VECTOR (15 downto 0) := X"0032";
constant c_COUNTER_WIDTH     : NATURAL := 16;
constant c_DATA_WIDTH        : NATURAL := 8;
constant c_RX_STACK_DEPTH    : NATURAL := 6;
constant c_TX_STACK_DEPTH    : NATURAL := 4;

-- fsm states of the i2c slave
type SLA_fsm is (S0_IDLE, S1_START_DETECT, S2_EVAL_FIRST_BYTE, S3_FIRST_ACKING, S4A_WRITE_SDA_BYTE, S4B_READ_SDA_BYTE, S5_FOLLOWING_ACKS, S6_PAUSE_DETECT, S7_ERROR, S8_WAIT_START_PAUSE);

signal i2c_SLA_fsm       : SLA_fsm;

type RX_FIFO_REG is array(0 to c_RX_STACK_DEPTH -1) of STD_LOGIC_VECTOR(c_DATA_WIDTH -1 downto 0);
type TX_FIFO_REG is array(0 to c_TX_STACK_DEPTH -1) of STD_LOGIC_VECTOR(c_DATA_WIDTH -1 downto 0);
 
-------------------------------------------------------------------------------
-- i2c signals 
-------------------------------------------------------------------------------
signal sda_oen_s    : STD_LOGIC;
signal sda_o_s     : STD_LOGIC;
signal scl_oen_s    : STD_LOGIC;

signal rx_fifo_dout  : RX_FIFO_REG;  -- This corresponds to the six DRXi registers

signal done_s      : STD_LOGIC;

signal start_o_s   : STD_LOGIC;
signal pause_o_s   : STD_LOGIC;
signal rcved_o_s   : STD_LOGIC;
signal caseCond            : STD_LOGIC_VECTOR(2 downto 0);

-- Bit counter signals
signal bit_sclr_s  : STD_LOGIC;
signal bit_count_s : STD_LOGIC_VECTOR(15 downto 0);
-- RX byte counter signals
signal rx_byte_sclr_s : STD_LOGIC;
signal rx_last_bit    : STD_LOGIC;
signal rx_byte_count_s: STD_LOGIC_VECTOR(15 downto 0);
-- TX byte counter signals
signal tx_byte_sclr_s : STD_LOGIC;
signal tx_last_bit    : STD_LOGIC;
signal tx_byte_count_s: STD_LOGIC_VECTOR(15 downto 0);

signal tx_byte        : STD_LOGIC_VECTOR(7 downto 0);

-- This is intended to be the personal address
constant i2c_general_addr   : STD_LOGIC_VECTOR (6 downto 0)  := "0000000";
signal i2c_addr             : STD_LOGIC_VECTOR (6 downto 0);
signal i2c_general_addr_OK  : BOOLEAN;
signal i2c_addr_OK          : BOOLEAN;

signal i2c_rd_wrn_inst      : STD_LOGIC;
signal ack_bit              : STD_LOGIC;

signal rx_fifo_din         : STD_LOGIC_VECTOR (c_DATA_WIDTH-1 downto 0);
signal rx_fifo_push        : STD_LOGIC;
signal rx_fifo_flush       : STD_LOGIC;

signal tx_fifo_oen_s       : STD_LOGIC;
signal tx_fifo_flush       : STD_LOGIC;
signal tx_fifo_reg_o       : STD_LOGIC_VECTOR (c_DATA_WIDTH -1 downto 0);    

signal bytes_ind_addr      : STD_LOGIC_VECTOR(7 downto 6);
signal bytes_wrn_inst      : STD_LOGIC_VECTOR (7 downto 0);
signal bytes_rd_inst       : STD_LOGIC_VECTOR (7 downto 0);

signal alarm_ind_wb_addr    : STD_LOGIC;
signal alarm_rd_inst_bytes  : STD_LOGIC;
signal alarm_wrn_inst_bytes : STD_LOGIC;


   component gc_counter
      generic (
         g_data_width : NATURAL := c_COUNTER_WIDTH
      );
      port (
         clk_i  : in  STD_LOGIC;
         rst_i  : in  STD_LOGIC;
         en_i   : in  STD_LOGIC;
         cnt_o  : out STD_LOGIC_VECTOR(g_data_width-1 downto 0)
      );
   end component;

   component i2c_bit
      generic(
         g_PRESCALER : STD_LOGIC_VECTOR (15 downto 0) := c_PRESCALER
      );
      port (
         rst_i     : in   STD_LOGIC;
         wb_clk_i  : in   STD_LOGIC;

         presc_e   : in   STD_LOGIC;
         presc     : in   STD_LOGIC_VECTOR (15 downto 0);

         done      : out  STD_LOGIC;
         sda_i     : in   STD_LOGIC;
         scl_i     : in   STD_LOGIC;
         start_o   : out  STD_LOGIC;
         pause_o   : out  STD_LOGIC;
         rcved_o   : out  STD_LOGIC
      );
   end component;

   component FIFO_stack
      generic(
         g_data_width  : NATURAL := c_DATA_WIDTH;
         g_stack_depth : NATURAL := c_RX_STACK_DEPTH
      );
      port (
         reg_i  : in   STD_LOGIC_VECTOR (g_data_width-1 downto 0);
         clk    : in   STD_LOGIC;
         push   : in   STD_LOGIC;
         flush  : in   STD_LOGIC;
         reg_o  : out  STD_LOGIC_VECTOR (g_stack_depth*g_data_width - 1 downto 0)
      );
   end component;

   component  FIFO_dispatcher
      generic(
         g_data_width       : NATURAL := c_DATA_WIDTH;
         g_dispatcher_depth : NATURAL := c_TX_STACK_DEPTH
      );
      port (
         reg_i  : in  STD_LOGIC_VECTOR (c_TX_STACK_DEPTH*c_DATA_WIDTH-1 downto 0);
         clk    : in  STD_LOGIC;
         load   : in  STD_LOGIC;
         flush  : in  STD_LOGIC;
         oen_i  : in  STD_LOGIC;
         reg_o  : out STD_LOGIC_VECTOR (c_DATA_WIDTH - 1 downto 0)
      );
   end component;

   component gc_ff
     port(
      Q   : out  STD_LOGIC;    
      C   : in   STD_LOGIC;
      CLR : in   STD_LOGIC;
      D   : in   STD_LOGIC
         );
   end component;

begin

caseCond              <= start_o_s & pause_o_s & rcved_o_s;
i2c_addr              <= CTR0_i (14 downto 8);
bytes_ind_addr        <= CTR0_i (7  downto 6);
bytes_wrn_inst        <= CTR1_i (15 downto 8);
bytes_rd_inst         <= CTR1_i (7  downto 0);

   bit_counter_16: gc_counter
   port map (
      clk_i  => wb_clk,
      rst_i  => bit_sclr_s or wb_rst_i,
      en_i   => done_s,
      cnt_o  => bit_count_s
   );

   rx_byte_counter_16: gc_counter
   port map (
      clk_i  => wb_clk,
      rst_i  => rx_byte_sclr_s or wb_rst_i,
      en_i   => rx_last_bit,
      cnt_o  => rx_byte_count_s
   );

   tx_byte_counter_16: gc_counter
   port map (
      clk_i  => wb_clk,
      rst_i  => tx_byte_sclr_s or wb_rst_i,
      en_i   => tx_last_bit,
      cnt_o  => tx_byte_count_s
   );

   inst_i2c_bit  : i2c_bit
   port map(
     rst_i     => wb_rst_i,
     wb_clk_i  => wb_clk,  -- Targeted at 20 MHz
     presc_e   => CTR0_i(2),
     presc     => PRE_i,
     done      => done_s,
     sda_i     => sda_i,
     scl_i     => scl_i,
     start_o   => start_o_s,
     pause_o   => pause_o_s,
     rcved_o   => rcved_o_s
   );

   -- We left half a wishbone cycle for setting the data before
   -- passing it to the FIFO

   ff_i2c_bit0: gc_ff
   port map(
      Q     => rx_fifo_din(0),
      C     => not(wb_clk) and done_s,
      CLR   => wb_rst_i,
      D     => rcved_o_s
   );

   ff_i2c_byte: for i in 1 to 7 generate
      ff_i2c_byte: gc_ff
      port map(
         Q     => rx_fifo_din(i),
         C     => not(wb_clk) and done_s,
         CLR   => wb_rst_i,
         D     => rx_fifo_din(i - 1)
      );
   end generate ff_i2c_byte;


   rx_fifo_i2c: FIFO_stack 
   port map(
      reg_i   => rx_fifo_din,
      clk     => wb_clk,
      push    => rx_last_bit,
      flush   => wb_rst_i or rx_fifo_flush,
      reg_o(7 downto 0)    => DRX0_o,
      reg_o(15 downto 8)   => DRX1_o, 
      reg_o(23 downto 16)  => DRX2_o, 
      reg_o(31 downto 24)  => DRX3_o, 
      reg_o(39 downto 32)  => DRX4_o, 
      reg_o(47 downto 40)  => DRX5_o  
   );

   tx_fifo_i2c: FIFO_dispatcher
   port map(
      reg_i(7  downto  0)  => DTX3_i,
      reg_i(15 downto  8)  => DTX2_i,
      reg_i(23 downto 16)  => DTX1_i,
      reg_i(31 downto 24)  => DTX0_i,
      clk     => wb_clk,
      load    => load_TX,
      flush   => tx_fifo_flush,
      oen_i   => tx_fifo_oen OR tx_fifo_oen_s,
      reg_o   => tx_fifo_reg_o
   );



reg_proc: process(wb_clk)


   procedure write_SDA is
   begin
      sda_oen_s       <= '0';
   end procedure;


   procedure read_SDA is
   begin
      sda_oen_s       <= '1';
   end procedure;


   procedure reset_proc is
   begin
      -- First we reset the fsm
      i2c_SLA_fsm    <= S0_IDLE;
      read_SDA;
      sda_o_s        <= '1';
      scl_oen_s       <= '1';
      -- By default we do not acknowledge
      ack_bit        <= '1';
      -- RX FIFO 
      rx_last_bit    <= '0';
      rx_fifo_flush  <= '0';
      i2c_rd_wrn_inst <= '0';
      -- Associated RX FIFO counter
      rx_byte_sclr_s       <= '0';
      -- TX FIFO
      tx_last_bit          <= '0';
      tx_fifo_flush        <= '0';
      -- Associated TX FIFO counter
      tx_byte_sclr_s       <= '0';
      -- Byte that it is being transmitted
      tx_byte              <= (others => '0');
      -- Alarm conditions are resetted
      alarm_ind_wb_addr    <= '0';
      alarm_rd_inst_bytes  <= '0';
      alarm_wrn_inst_bytes <= '0';
   end procedure;


   procedure eval_byte is
   begin
   case i2c_SLA_fsm is

         when S1_START_DETECT =>
            i2c_SLA_fsm      <= S2_EVAL_FIRST_BYTE;
            rx_byte_sclr_s   <= '0';
            rx_fifo_flush    <= '0';
            tx_byte_sclr_s   <= '0';
            tx_fifo_flush    <= '0';
         when S2_EVAL_FIRST_BYTE =>
            -- Here the count goes with one less value, due to the effect of 
            -- the start condition while counting:
            -- we can only tell once it happened so that there's one delay
            -- in the count of the bits for the first byte.
            case bit_count_s is
            when X"0006" =>
               if rx_fifo_din(6 downto 0) = i2c_general_addr or rx_fifo_din(6 downto 0) = i2c_addr then
                  ack_bit <= '0';
               else
                  ack_bit <= '1';
               end if;
            when X"0007" =>
               i2c_SLA_fsm       <= S3_FIRST_ACKING;
               i2c_rd_wrn_inst   <= rx_fifo_din(0);
               -- According to the specification, the maximum delay between the
               -- SCL falling edge and the placement of the ACK in the bus 
               -- should be less than 31% of the period. So with one eigth of
               -- it should be ok.
               -- However, as we already have some delays in the deglitchers,
               -- we will not place that timer.
               write_SDA;
               sda_o_s           <= ack_bit;
               bit_sclr_s        <= '1';
            when others =>
            end case;
         when S4B_READ_SDA_BYTE =>
            case bit_count_s is
            when X"0008" =>
               i2c_SLA_fsm <= S5_FOLLOWING_ACKS;
               write_SDA;
               ack_bit       <= '0';
               sda_o_s       <= '0';
               -- Push the FIFO_stack one position
               rx_last_bit   <= '1';
               bit_sclr_s    <= '1';

               if (rx_byte_count_s = X"000" & "00" & bytes_ind_addr) then
                 alarm_ind_wb_addr <= '1';
               end if;

            when others =>
            end case;
         when others =>
         end case;

   end procedure;

	
   procedure check_rcved_data is
   begin
      if sda_oen_s = '1' then
      case caseCond is
      -- This is the start condition
      when "101" =>
         i2c_SLA_fsm      <= S1_START_DETECT;
         bit_sclr_s       <= '1';
         rx_byte_sclr_s   <= '1';
         tx_byte_sclr_s   <= '1';
         -- Check for indirect read
         -- TODO: the generation of alarm_ind_wb_addr should be erased from
         -- here. BUT do not really know what to do with the flush command.
         if (rx_byte_count_s = X"000"&"00"&bytes_ind_addr ) then
            alarm_ind_wb_addr <= '1';
         else
            rx_fifo_flush <= '1';
         end if;
         -- This corresponds to the stop condition
      when "010" =>
         i2c_SLA_fsm      <= S6_PAUSE_DETECT;
         bit_sclr_s       <= '1';
         rx_byte_sclr_s   <= '1';
         tx_byte_sclr_s   <= '1';
      -- The reception of a high and low bit
      when "000" =>
         eval_byte;
      when "001" =>
         eval_byte;
      -- Same code as "100"
      when "110" =>
         i2c_SLA_fsm  <= S7_ERROR;
      when "111" =>
         i2c_SLA_fsm  <= S7_ERROR;
      when others =>
         i2c_SLA_fsm  <= S0_IDLE;
      end case;
    else
    end if;
   end procedure;


begin

   if rising_edge(wb_clk) then
      if wb_rst_i = '1' then
         reset_proc;
      else
         -- Update all outputs
         sda_oen       <= sda_oen_s;
         sda_o        <= sda_o_s;
         scl_oen       <= scl_oen_s;

         if done_s = '1' then
            case i2c_SLA_fsm is

            when S0_IDLE            =>
               check_rcved_data;

            when S1_START_DETECT    =>
               check_rcved_data;

            when S2_EVAL_FIRST_BYTE =>
               check_rcved_data;

            when S3_FIRST_ACKING    =>
               case caseCond is
               when "101" =>
                  i2c_SLA_fsm  <= S1_START_DETECT;
                  read_SDA;
                  sda_o_s         <= '1';
                  ack_bit         <= '1';
                  -- Here we prepare bit_counter_16 that will be deresetted next fsm
                  bit_sclr_s      <= '1';
                  rx_byte_sclr_s  <= '1';
                  tx_byte_sclr_s  <= '1';
               when "010" =>
                  i2c_SLA_fsm     <= S6_PAUSE_DETECT;
                  read_SDA;
                  bit_sclr_s      <= '1';
                  rx_byte_sclr_s  <= '1';
                  tx_byte_sclr_s  <= '1';
                  ack_bit         <= '1';
                  sda_o_s         <= '1';
               when others => 
                  case i2c_rd_wrn_inst is
                  when '1'    =>
                     if ack_bit = '0' then
                        i2c_SLA_fsm   <= S4A_WRITE_SDA_BYTE;
                        write_SDA;
                        ack_bit       <= '1';
                        -- Start sending the first byte
                        sda_o_s       <= tx_fifo_reg_o(0);
                        tx_byte       <= '0' & tx_fifo_reg_o(7 downto 1);
                     else
                        i2c_SLA_fsm <= S8_WAIT_START_PAUSE; 
                     end if;
                  when others  =>
                     read_SDA;
                     if ack_bit = '0' then
                        i2c_SLA_fsm    <= S4B_READ_SDA_BYTE;
                        ack_bit        <= '1';
                        sda_o_s        <= '1';
                     else
                        i2c_SLA_fsm <= S8_WAIT_START_PAUSE; 
                     end if;
                  end case;
               end case;

            when S4A_WRITE_SDA_BYTE =>
               case bit_count_s is
               when X"0008"     =>
                  i2c_SLA_fsm   <= S5_FOLLOWING_ACKS; 
                  read_SDA;
                  sda_o_s       <= '0';
                  ack_bit       <= '0';
                  tx_last_bit   <= '1';
                  bit_sclr_s    <= '1'; 
                  -- Shift one position the FIFO_dispatcher
                  tx_fifo_oen_s <= '1';
               when others =>
                  sda_o_s       <= tx_byte(0);
                  tx_byte       <= '0' & tx_byte(7 downto 1);
               end case;

            when S4B_READ_SDA_BYTE  =>
               check_rcved_data;

            when S5_FOLLOWING_ACKS =>
               tx_last_bit   <= '0';
               bit_sclr_s    <= '0';
               sda_o_s       <= '1';
               case i2c_rd_wrn_inst is
               when '1'            =>
                 if tx_byte_count_s(7 downto 0) =  bytes_rd_inst then
                   i2c_SLA_fsm     <= S8_WAIT_START_PAUSE;
                   read_SDA;
                   ack_bit         <= '1';
                   tx_byte_sclr_s  <= '1';
                   alarm_rd_inst_bytes <= '1';
                 else
                   i2c_SLA_fsm     <= S4A_WRITE_SDA_BYTE;
                   write_SDA;
                   ack_bit         <= '1';
                   -- Send the following byte in memory
                   sda_o_s         <= tx_fifo_reg_o(0);
                   tx_byte         <= '0' & tx_fifo_reg_o(7 downto 1);
                 end if;
               when others         =>
                  read_SDA;
                  if rx_byte_count_s(7 downto 0) = bytes_wrn_inst then
                     i2c_SLA_fsm            <= S8_WAIT_START_PAUSE;
                     rx_byte_sclr_s         <= '1';
                     alarm_wrn_inst_bytes   <= '1';
                  else
                     i2c_SLA_fsm      <= S4B_READ_SDA_BYTE;
                     ack_bit          <= '1';
                  end if;
                  check_rcved_data;
               end case;

            when S6_PAUSE_DETECT     =>
               check_rcved_data;

            when S8_WAIT_START_PAUSE =>
               check_rcved_data;

            when others              =>
            end case;

         else
            bit_sclr_s             <= '0';
            rx_byte_sclr_s         <= '0';
            tx_byte_sclr_s         <= '0';

            alarm_rd_inst_bytes    <= '0';
            alarm_wrn_inst_bytes   <= '0';
            alarm_ind_wb_addr      <= '0';

            tx_last_bit            <= '0';
            tx_fifo_oen_s          <= '0';
            
            rx_last_bit            <= '0';
         end if;
      end if;
   else
   end if;
end process;


end Behavioral;
